@@ -1,5 +1,19 @@
 import types
 
+try:
+    callable
+except NameError:
+    # Python 3.1
+    from collections import Callable
+    callable = lambda x: isinstance(x, Callable)
+
+try:
+    basestring
+except NameError:
+    # Python 3.x
+    basestring = str
+
+
 class FFIError(Exception):
     pass
 
@@ -39,6 +53,8 @@ class FFI(object):
             # You need PyPy (>= 2.0 beta), or a CPython (>= 2.6) with
             # _cffi_backend.so compiled.
             import _cffi_backend as backend
+            from . import __version__
+            assert backend.__version__ == __version__
             # (If you insist you can also try to pass the option
             # 'backend=backend_ctypes.CTypesBackend()', but don't
             # rely on it!  It's probably not going to work well.)
@@ -57,16 +73,15 @@ class FFI(object):
             if name.startswith('RTLD_'):
                 setattr(self, name, getattr(backend, name))
         #
-        self._parser._declarations['typedef FILE'] = model.file_type
-        BVoidP = self._get_cached_btype(model.voidp_type)
+        self.BVoidP = self._get_cached_btype(model.voidp_type)
         if isinstance(backend, types.ModuleType):
             # _cffi_backend: attach these constants to the class
             if not hasattr(FFI, 'NULL'):
-                FFI.NULL = self.cast(BVoidP, 0)
+                FFI.NULL = self.cast(self.BVoidP, 0)
                 FFI.CData, FFI.CType = backend._get_types()
         else:
             # ctypes backend: attach these constants to the instance
-            self.NULL = self.cast(BVoidP, 0)
+            self.NULL = self.cast(self.BVoidP, 0)
             self.CData, self.CType = backend._get_types()
 
     def cdef(self, csource, override=False):
@@ -75,6 +90,10 @@ class FFI(object):
         then be accessed via either 'ffi.dlopen()' or 'ffi.verify()'.
         The types can be used in 'ffi.new()' and other functions.
         """
+        if not isinstance(csource, str):    # unicode, on Python 2
+            if not isinstance(csource, basestring):
+                raise TypeError("cdef() argument must be a string")
+            csource = csource.encode('ascii')
         self._parser.parse(csource, override=override)
         self._cdefsources.append(csource)
         if override:
@@ -88,7 +107,7 @@ class FFI(object):
         linked to a particular library, just like C headers; in the
         library we only look for the actual (untyped) symbols.
         """
-        assert isinstance(name, str) or name is None
+        assert isinstance(name, basestring) or name is None
         lib, function_cache = _make_ffi_library(self, name, flags)
         self._function_caches.append(function_cache)
         self._libraries.append(lib)
@@ -101,11 +120,14 @@ class FFI(object):
             if consider_function_as_funcptr and not cfaf:
                 raise KeyError
         except KeyError:
+            key = cdecl
+            if not isinstance(cdecl, str):    # unicode, on Python 2
+                cdecl = cdecl.encode('ascii')
             cfaf = consider_function_as_funcptr
             type = self._parser.parse_type(cdecl,
                        consider_function_as_funcptr=cfaf)
             btype = self._get_cached_btype(type)
-            self._parsed_types[cdecl] = btype, cfaf
+            self._parsed_types[key] = btype, cfaf
         return btype
 
     def typeof(self, cdecl):
@@ -113,16 +135,21 @@ class FFI(object):
         corresponding Python type: <class 'ffi.CData<...>'>.
         It can also be used on 'cdata' instance to get its C type.
         """
-        if isinstance(cdecl, str):
+        if isinstance(cdecl, basestring):
             return self._typeof(cdecl)
-        else:
+        if isinstance(cdecl, self.CData):
             return self._backend.typeof(cdecl)
+        if isinstance(cdecl, types.BuiltinFunctionType):
+            res = _builtin_function_type(cdecl)
+            if res is not None:
+                return res
+        raise TypeError(type(cdecl))
 
     def sizeof(self, cdecl):
         """Return the size in bytes of the argument.  It can be a
         string naming a C type, or a 'cdata' instance.
         """
-        if isinstance(cdecl, str):
+        if isinstance(cdecl, basestring):
             BType = self._typeof(cdecl)
             return self._backend.sizeof(BType)
         else:
@@ -132,7 +159,7 @@ class FFI(object):
         """Return the natural alignment size in bytes of the C type
         given as a string.
         """
-        if isinstance(cdecl, str):
+        if isinstance(cdecl, basestring):
             cdecl = self._typeof(cdecl)
         return self._backend.alignof(cdecl)
 
@@ -140,7 +167,7 @@ class FFI(object):
         """Return the offset of the named field inside the given
         structure, which must be given as a C type name.
         """
-        if isinstance(cdecl, str):
+        if isinstance(cdecl, basestring):
             cdecl = self._typeof(cdecl)
         return self._backend.typeoffsetof(cdecl, fieldname)[1]
 
@@ -167,7 +194,7 @@ class FFI(object):
         about that when copying the pointer to the memory somewhere
         else, e.g. into another structure.
         """
-        if isinstance(cdecl, str):
+        if isinstance(cdecl, basestring):
             cdecl = self._typeof(cdecl)
         return self._backend.newp(cdecl, init)
 
@@ -176,7 +203,7 @@ class FFI(object):
         type initialized with the given 'source'.  The source is
         casted between integers or pointers of any type.
         """
-        if isinstance(cdecl, str):
+        if isinstance(cdecl, basestring):
             cdecl = self._typeof(cdecl)
         return self._backend.cast(cdecl, source)
 
@@ -194,7 +221,7 @@ class FFI(object):
         it as a string or unicode string.
 
         If 'cdata' is an enum, returns the value of the enumerator as a
-        string, or '#NUMBER' if the value is out of range.
+        string, or 'NUMBER' if the value is out of range.
         """
         return self._backend.string(cdata, maxlen)
 
@@ -224,7 +251,7 @@ class FFI(object):
                 raise TypeError("the 'python_callable' argument "
                                 "is not callable")
             return self._backend.callback(cdecl, python_callable, error)
-        if isinstance(cdecl, str):
+        if isinstance(cdecl, basestring):
             cdecl = self._typeof(cdecl, consider_function_as_funcptr=True)
         if python_callable is None:
             return callback_decorator_wrap                # decorator mode
@@ -237,7 +264,7 @@ class FFI(object):
         extra text to append (or insert for more complicated C types), like
         a variable name, or '*' to get actually the C type 'pointer-to-cdecl'.
         """
-        if isinstance(cdecl, str):
+        if isinstance(cdecl, basestring):
             cdecl = self._typeof(cdecl)
         replace_with = replace_with.strip()
         if (replace_with.startswith('*')
@@ -305,30 +332,54 @@ class FFI(object):
         ctypeptr = self._pointer_to(ctype)
         return self._backend.rawaddressof(ctypeptr, cdata, offset)
 
+    def include(self, ffi_to_include):
+        """Includes the typedefs, structs, unions and enums defined
+        in another FFI instance.  Usage is similar to a #include in C,
+        where a part of the program might include types defined in
+        another part for its own usage.  Note that the include()
+        method has no effect on functions, constants and global
+        variables, which must anyway be accessed directly from the
+        lib object returned by the original FFI instance.
+        """
+        self._parser.include(ffi_to_include._parser)
+        self._cdefsources.append('[')
+        self._cdefsources.extend(ffi_to_include._cdefsources)
+        self._cdefsources.append(']')
+
+    def new_handle(self, x):
+        return self._backend.newp_handle(self.BVoidP, x)
+
+    def from_handle(self, x):
+        return self._backend.from_handle(x)
+
 
 def _make_ffi_library(ffi, libname, flags):
     import os
     name = libname
     if name is None:
         name = 'c'    # on Posix only
-    if os.path.sep in name or (
-            os.path.altsep is not None and os.path.altsep in name):
-        path = name
-    else:
+    backend = ffi._backend
+    try:
+        if '.' not in name and '/' not in name:
+            raise OSError("library not found: %r" % (name,))
+        backendlib = backend.load_library(name, flags)
+    except OSError:
         import ctypes.util
         path = ctypes.util.find_library(name)
         if path is None:
-            raise OSError("library not found: %r" % (name,))
-    #
-    backend = ffi._backend
-    backendlib = backend.load_library(path, flags)
+            raise     # propagate the original OSError
+        backendlib = backend.load_library(path, flags)
+    copied_enums = []
     #
     def make_accessor(name):
         key = 'function ' + name
         if key in ffi._parser._declarations:
             tp = ffi._parser._declarations[key]
             BType = ffi._get_cached_btype(tp)
-            value = backendlib.load_function(BType, name)
+            try:
+                value = backendlib.load_function(BType, name)
+            except KeyError as e:
+                raise AttributeError('%s: %s' % (name, e))
             library.__dict__[name] = value
             return
         #
@@ -343,6 +394,18 @@ def _make_ffi_library(ffi, libname, flags):
                 lambda self, value: write_variable(BType, name, value)))
             return
         #
+        if not copied_enums:
+            from . import model
+            for key, tp in ffi._parser._declarations.items():
+                if not isinstance(tp, model.EnumType):
+                    continue
+                for enumname, enumval in zip(tp.enumerators, tp.enumvalues):
+                    if enumname not in library.__dict__:
+                        library.__dict__[enumname] = enumval
+            copied_enums.append(True)
+        #
+        if name in library.__dict__:   # copied from an enum value just above,
+            return                     # or multithread's race condition
         raise AttributeError(name)
     #
     class FFILibrary(object):
@@ -359,6 +422,25 @@ def _make_ffi_library(ffi, libname, flags):
                 property.__set__(self, value)
     #
     if libname is not None:
-        FFILibrary.__name__ = 'FFILibrary_%s' % libname
+        try:
+            if not isinstance(libname, str):    # unicode, on Python 2
+                libname = libname.encode('utf-8')
+            FFILibrary.__name__ = 'FFILibrary_%s' % libname
+        except UnicodeError:
+            pass
     library = FFILibrary()
     return library, library.__dict__
+
+def _builtin_function_type(func):
+    # a hack to make at least ffi.typeof(builtin_function) work,
+    # if the builtin function was obtained by 'vengine_cpy'.
+    import sys
+    try:
+        module = sys.modules[func.__module__]
+        ffi = module._cffi_original_ffi
+        types_of_builtin_funcs = module._cffi_types_of_builtin_funcs
+        tp = types_of_builtin_funcs[func]
+    except (KeyError, AttributeError, TypeError):
+        return None
+    else:
+        return ffi._get_cached_btype(tp)
